from typing import List, Literal, Optional

from pydantic import BaseModel, Field

from pyjlyric.model import LyricPage, WithUrlText


class MusicbookLyricPage(LyricPage):
    artist: WithUrlText
    composer: Optional[str]
    lyricist: Optional[str]
    arranger: None


class MusicbookLyricData(BaseModel):
    title: Optional[str]
    writer: Optional[str]
    composer: Optional[str]
    lyrics: List[str]
    error_code: Optional[Literal["apiCallError"]] = Field(..., alias="errorCode")



from base64 import standard_b64decode
from collections.abc import Generator
from typing import Dict, List, Optional

from pydantic import BaseModel, validator

from pyjlyric.model import LyricPage, WithUrlText


class PetitlyricsLyricPage(LyricPage):
    artist: WithUrlText
    composer: Optional[str]
    lyricist: Optional[str]
    arranger: None


class _PetitlyricsLyric(BaseModel):
    lyrics: str

    @validator("lyrics")
    @classmethod
    def validate_lyrics(cls, v: str) -> str:
        return standard_b64decode(v).decode("utf-8")


class PetitlyricsLyricData(BaseModel):
    __root__: List[str]

    def __iter__(self) -> Generator[str, None, None]:  # type: ignore[override]
        yield from self.__root__

    @validator("__root__", pre=True, each_item=True)
    @classmethod
    def validate_root(cls, v: Dict[str, str]) -> str:
        return _PetitlyricsLyric(**v).lyrics



"""<https://kashinavi.com>"""

import re

from bs4 import BeautifulSoup, NavigableString, Tag

from pyjlyric.base import BaseLyricPageParser, BaseLyricPageParserError
from pyjlyric.model import WithUrlText
from pyjlyric.util import get_captured_value, get_source, parse_obj_as_url

from .model import JSONLD, KashinaviLyricPage

_KASHINAVI_PATTERN = r"^https://kashinavi\.com/song_view\.html\?(?P<pageid>\d+)"


class KashinaviLyricPageParserError(BaseLyricPageParserError):
    """WIP."""


class KashinaviLyricPageParser(BaseLyricPageParser):
    """https://kashinavi.com/song_view.html?<pageid>"""

    _test = "https://kashinavi.com/song_view.html?155779"

    @staticmethod
    def is_valid_url(url: str) -> bool:
        """Check if the url is valid."""
        pattern = re.compile(_KASHINAVI_PATTERN)
        m = re.match(pattern, url)
        return get_captured_value(m, "pageid") is not None

    @staticmethod
    def parse(url: str) -> KashinaviLyricPage:
        """Parse the url page and return the result as LyricPage instance."""
        pattern = re.compile(_KASHINAVI_PATTERN)
        m = re.match(pattern, url)
        pageid = get_captured_value(m, "pageid")
        if pageid is None:
            raise KashinaviLyricPageParserError from ValueError

        bs = get_source(url)
        if bs is None:
            raise KashinaviLyricPageParserError from ConnectionError

        jsonld_tag: Tag = bs.find_all("script", type="application/ld+json")[1]

        jsonld_source = jsonld_tag.text
        jsonld = JSONLD.parse_raw(jsonld_source)

        artist = jsonld.recorded_as.by_artist

        return KashinaviLyricPage(
            title=jsonld.name,
            page_url=parse_obj_as_url(url),
            pageid=pageid,
            artist=WithUrlText(link=artist.url, text=artist.name),
            composer=jsonld.composer.name,
            lyricist=jsonld.lyricist.name,
            arranger=None,
            lyric_sections=[
                [i.text for i in section.children if isinstance(i, NavigableString)]
                for section in BeautifulSoup(jsonld.lyrics.text, "lxml").find_all("p")
            ],
        )
